@@ -14,7 +14,7 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-set -e
+set -e -o pipefail
 
 _testlib_this="${BASH_SOURCE[0]}"
 _testlib_dir="$( cd "$( dirname "${BASH_SOURCE[0]}" )" >/dev/null 2>&1 && pwd )"
@@ -39,31 +39,36 @@ create_results_dir() {
   chmod ogu+w "$RESULT_DIR"
 }
 
-## @description find all the test.sh scripts in the immediate child dirs
+## @description find all the test*.sh scripts in the immediate child dirs
 all_tests_in_immediate_child_dirs() {
-  find . -mindepth 2 -maxdepth 2 -name test.sh | cut -c3- | sort
+  find . -mindepth 2 -maxdepth 2 -name 'test*.sh' | cut -c3- | sort
 }
 
-## @description Find all test.sh scripts in immediate child dirs,
+## @description Find all test*.sh scripts in immediate child dirs,
 ## @description applying OZONE_ACCEPTANCE_SUITE or OZONE_TEST_SELECTOR filter.
 find_tests(){
   if [[ -n "${OZONE_ACCEPTANCE_SUITE}" ]]; then
-     tests=$(all_tests_in_immediate_child_dirs | xargs grep -l "^#suite:${OZONE_ACCEPTANCE_SUITE}$")
+    tests=$(all_tests_in_immediate_child_dirs | xargs grep -l "^#suite:${OZONE_ACCEPTANCE_SUITE}$" || echo "")
 
      # 'misc' is default suite, add untagged tests, too
     if [[ "misc" == "${OZONE_ACCEPTANCE_SUITE}" ]]; then
-       untagged="$(all_tests_in_immediate_child_dirs | xargs grep -L "^#suite:")"
-       if [[ -n "${untagged}" ]]; then
-         tests=$(echo ${tests} ${untagged} | xargs -n1 | sort)
-       fi
-     fi
+      untagged="$(all_tests_in_immediate_child_dirs | xargs grep -L "^#suite:")"
+      if [[ -n "${untagged}" ]]; then
+        tests=$(echo ${tests} ${untagged} | xargs -n1 | sort)
+      fi
+    fi
 
     if [[ -z "${tests}" ]]; then
-       echo "No tests found for suite ${OZONE_ACCEPTANCE_SUITE}"
-       exit 1
+      echo "No tests found for suite ${OZONE_ACCEPTANCE_SUITE}"
+      exit 1
     fi
   elif [[ -n "${OZONE_TEST_SELECTOR}" ]]; then
-    tests=$(all_tests_in_immediate_child_dirs | grep "${OZONE_TEST_SELECTOR}")
+    tests=$(all_tests_in_immediate_child_dirs | grep "${OZONE_TEST_SELECTOR}" || echo "")
+
+    if [[ -z "${tests}" ]]; then
+      echo "No tests found for filter ${OZONE_TEST_SELECTOR}"
+      exit 1
+    fi
   else
     tests=$(all_tests_in_immediate_child_dirs | xargs grep -L '^#suite:failing')
   fi
@@ -72,38 +77,14 @@ find_tests(){
 
 ## @description wait until safemode exit (or 240 seconds)
 wait_for_safemode_exit(){
-  # version-dependent
-  : ${OZONE_SAFEMODE_STATUS_COMMAND:=ozone admin safemode status --verbose}
+  local cmd="ozone admin safemode wait -t 240"
+  if [[ "${SECURITY_ENABLED}" == 'true' ]]; then
+    wait_for_port kdc 88 60
+    cmd="kinit -k HTTP/scm@EXAMPLE.COM -t /etc/security/keytabs/HTTP.keytab && $cmd"
+  fi
 
-  #Reset the timer
-  SECONDS=0
-
-  #Don't give it up until 240 seconds
-  while [[ $SECONDS -lt 240 ]]; do
-
-     #This line checks the safemode status in scm
-     local command="${OZONE_SAFEMODE_STATUS_COMMAND}"
-     if [[ "${SECURITY_ENABLED}" == 'true' ]]; then
-         status=$(docker-compose exec -T ${SCM} bash -c "kinit -k HTTP/scm@EXAMPLE.COM -t /etc/security/keytabs/HTTP.keytab && $command" || true)
-     else
-         status=$(docker-compose exec -T ${SCM} bash -c "$command")
-     fi
-
-     echo "SECONDS: $SECONDS"
-
-     echo $status
-     if [[ "$status" ]]; then
-       if [[ ${status} == "SCM is out of safe mode." ]]; then
-         #Safemode exits. Let's return from the function.
-         echo "Safe mode is off"
-         return
-       fi
-     fi
-
-     sleep 2
-   done
-   echo "WARNING! Safemode is still on. Please check the docker-compose files"
-   return 1
+  wait_for_port ${SCM} 9860 120
+  execute_commands_in_container ${SCM} "$cmd"
 }
 
 ## @description wait until OM leader is elected (or 120 seconds)
@@ -120,12 +101,22 @@ wait_for_om_leader() {
   while [[ $SECONDS -lt 120 ]]; do
     local command="ozone admin om getserviceroles --service-id '${OM_SERVICE_ID}'"
     if [[ "${SECURITY_ENABLED}" == 'true' ]]; then
-      status=$(docker-compose exec -T ${SCM} bash -c "kinit -k scm/scm@EXAMPLE.COM -t /etc/security/keytabs/scm.keytab && $command" | grep LEADER)
+      status=$(docker-compose exec -T ${SCM} bash -c "kinit -k scm/scm@EXAMPLE.COM -t /etc/security/keytabs/scm.keytab && $command" | grep LEADER || true)
     else
-      status=$(docker-compose exec -T ${SCM} bash -c "$command" | grep LEADER)
+      status=$(docker-compose exec -T ${SCM} bash -c "$command" | grep LEADER || true)
     fi
     if [[ -n "${status}" ]]; then
       echo "Found OM leader for service ${OM_SERVICE_ID}: $status"
+
+      local grep_command="grep -e FOLLOWER -e LEADER | sort -r -k3 | awk '{ print \$1 }' | xargs echo | sed 's/ /,/g'"
+      local new_order
+      if [[ "${SECURITY_ENABLED}" == 'true' ]]; then
+        new_order=$(docker-compose exec -T ${SCM} bash -c "kinit -k scm/scm@EXAMPLE.COM -t /etc/security/keytabs/scm.keytab && $command | ${grep_command}")
+      else
+        new_order=$(docker-compose exec -T ${SCM} bash -c "$command | ${grep_command}")
+      fi
+
+      reorder_om_nodes "${new_order}"
       return
     else
       echo "Waiting for OM leader for service ${OM_SERVICE_ID}"
@@ -147,14 +138,13 @@ start_docker_env(){
   create_results_dir
   export OZONE_SAFEMODE_MIN_DATANODES="${datanode_count}"
 
-  docker-compose --no-ansi down
-  if ! { docker-compose --no-ansi up -d --scale datanode="${datanode_count}" \
-      && wait_for_safemode_exit \
-      && wait_for_om_leader ; }; then
-    [[ -n "$OUTPUT_NAME" ]] || OUTPUT_NAME="$COMPOSE_ENV_NAME"
-    stop_docker_env
-    return 1
-  fi
+  docker-compose --ansi never down
+
+  trap stop_docker_env EXIT HUP INT TERM
+
+  docker-compose --ansi never up -d --scale datanode="${datanode_count}"
+  wait_for_safemode_exit
+  wait_for_om_leader
 }
 
 ## @description  Execute robot tests in a specific container.
@@ -169,20 +159,23 @@ execute_robot_test(){
   unset 'ARGUMENTS[${#ARGUMENTS[@]}-1]' #Remove the last element, remainings are the custom parameters
   TEST_NAME=$(basename "$TEST")
   TEST_NAME="$(basename "$COMPOSE_DIR")-${TEST_NAME%.*}"
-  set +e
-  [[ -n "$OUTPUT_NAME" ]] || OUTPUT_NAME="$COMPOSE_ENV_NAME-$TEST_NAME-$CONTAINER"
+
+  local output_name=$(get_output_name)
 
   # find unique filename
   declare -i i=0
-  OUTPUT_FILE="robot-${OUTPUT_NAME}.xml"
+  OUTPUT_FILE="robot-${output_name}1.xml"
   while [[ -f $RESULT_DIR/$OUTPUT_FILE ]]; do
-    let i++
-    OUTPUT_FILE="robot-${OUTPUT_NAME}-${i}.xml"
+    let ++i
+    OUTPUT_FILE="robot-${output_name}${i}.xml"
   done
 
   SMOKETEST_DIR_INSIDE="${OZONE_DIR:-/opt/hadoop}/smoketest"
 
   OUTPUT_PATH="$RESULT_DIR_INSIDE/${OUTPUT_FILE}"
+
+  set +e
+
   # shellcheck disable=SC2068
   docker-compose exec -T "$CONTAINER" mkdir -p "$RESULT_DIR_INSIDE" \
     && docker-compose exec -T "$CONTAINER" robot \
@@ -199,22 +192,44 @@ execute_robot_test(){
   FULL_CONTAINER_NAME=$(docker-compose ps | grep "_${CONTAINER}_" | head -n 1 | awk '{print $1}')
   docker cp "$FULL_CONTAINER_NAME:$OUTPUT_PATH" "$RESULT_DIR/"
 
-  copy_daemon_logs
+  if [[ ${rc} -gt 0 ]] && [[ ${rc} -le 250 ]]; then
+    create_stack_dumps
+  fi
 
   set -e
 
-  if [[ ${rc} -gt 0 ]]; then
-    stop_docker_env
-  fi
-
   return ${rc}
 }
 
+## @description Replace OM node order in config
+reorder_om_nodes() {
+  local c pid procname new_order
+  local new_order="$1"
+
+  if [[ -n "${new_order}" ]] && [[ "${new_order}" != "om1,om2,om3" ]]; then
+    for c in $(docker-compose ps | cut -f1 -d' ' | grep -e datanode -e recon -e s3g -e scm); do
+      docker exec "${c}" sed -i -e "s/om1,om2,om3/${new_order}/" /etc/hadoop/ozone-site.xml
+      echo "Replaced OM order with ${new_order} in ${c}"
+    done
+  fi
+}
+
+## @description Create stack dump of each java process in each container
+create_stack_dumps() {
+  local c pid procname
+  for c in $(docker-compose ps | cut -f1 -d' ' | grep -e datanode -e om -e recon -e s3g -e scm); do
+    while read -r pid procname; do
+      echo "jstack $pid > ${RESULT_DIR}/${c}_${procname}.stack"
+      docker exec "${c}" bash -c "jstack $pid" > "${RESULT_DIR}/${c}_${procname}.stack"
+    done < <(docker exec "${c}" sh -c "jps | grep -v Jps" || true)
+  done
+}
+
 ## @description Copy any 'out' files for daemon processes to the result dir
 copy_daemon_logs() {
   local c f
   for c in $(docker-compose ps | grep "^${COMPOSE_ENV_NAME}_" | awk '{print $1}'); do
-    for f in $(docker exec "${c}" ls -1 /var/log/hadoop | grep -F '.out'); do
+    for f in $(docker exec "${c}" ls -1 /var/log/hadoop 2> /dev/null | grep -F -e '.out' -e audit); do
       docker cp "${c}:/var/log/hadoop/${f}" "$RESULT_DIR/"
     done
   done
@@ -225,27 +240,51 @@ copy_daemon_logs() {
 ## @param        container name
 ## @param        specific command to execute
 execute_command_in_container(){
-  set -e
   # shellcheck disable=SC2068
   docker-compose exec -T "$@"
-  set +e
+}
+
+## @description  Execute specific commands in docker container
+## @param        container name
+## @param        specific commands to execute
+execute_commands_in_container(){
+  local container=$1
+  shift 1
+  local command=$@
+
+  # shellcheck disable=SC2068
+  docker-compose exec -T $container /bin/bash -c "$command"
 }
 
 ## @description Stop a list of named containers
 ## @param       List of container names, eg datanode_1 datanode_2
 stop_containers() {
-  set -e
-  docker-compose --no-ansi stop $@
-  set +e
+  docker-compose --ansi never stop $@
 }
 
 
 ## @description Start a list of named containers
 ## @param       List of container names, eg datanode_1 datanode_2
 start_containers() {
-  set -e
-  docker-compose --no-ansi start $@
-  set +e
+  docker-compose --ansi never start $@
+}
+
+create_containers() {
+  docker-compose --ansi never up -d $@
+}
+
+get_output_name() {
+  if [[ -n "${OUTPUT_NAME}" ]]; then
+    echo "${OUTPUT_NAME}-"
+  fi
+}
+
+save_container_logs() {
+  local output_name=$(get_output_name)
+  local c
+  for c in $(docker-compose ps "$@" | cut -f1 -d' ' | tail -n +3); do
+    docker logs "${c}" >> "$RESULT_DIR/docker-${output_name}${c}.log" 2>&1
+  done
 }
 
 
@@ -262,13 +301,9 @@ wait_for_port(){
   SECONDS=0
 
   while [[ $SECONDS -lt $timeout ]]; do
-     set +e
-     docker-compose exec -T ${SCM} /bin/bash -c "nc -z $host $port"
-     status=$?
-     set -e
-     if [ $status -eq 0 ] ; then
-         echo "Port $port is available on $host"
-         return;
+     if docker-compose exec -T ${SCM} /bin/bash -c "nc -z $host $port"; then
+       echo "Port $port is available on $host"
+       return
      fi
      echo "Port $port is not available on $host yet"
      sleep 1
@@ -277,12 +312,50 @@ wait_for_port(){
    return 1
 }
 
+## @description wait for the stat to be ready
+## @param The container ID
+## @param The maximum time to wait in seconds
+## @param The command line to be executed
+wait_for_execute_command(){
+  local container=$1
+  local timeout=$2
+  shift 2
+  local command=$@
+
+  #Reset the timer
+  SECONDS=0
+
+  while [[ $SECONDS -lt $timeout ]]; do
+     if docker-compose exec -T $container /bin/bash -c "$command"; then
+        echo "$command succeed"
+        return
+     fi
+        echo "$command hasn't succeed yet"
+        sleep 1
+   done
+   echo "Timed out waiting on $command to be successful"
+   return 1
+}
 
 ## @description  Stops a docker-compose based test environment (with saving the logs)
 stop_docker_env(){
-  docker-compose --no-ansi logs > "$RESULT_DIR/docker-$OUTPUT_NAME.log"
+  copy_daemon_logs
+  save_container_logs
   if [ "${KEEP_RUNNING:-false}" = false ]; then
-     docker-compose --no-ansi down
+    down_repeats=3
+    for i in $(seq 1 $down_repeats)
+    do
+      if docker-compose --ansi never down; then
+        return
+      fi
+      if [[ ${i} -eq 1 ]]; then
+        create_stack_dumps
+      fi
+      sleep 5
+    done
+
+    echo "Failed to remove all docker containers in $down_repeats attempts."
+    return 1
   fi
 }
 
@@ -316,36 +389,44 @@ generate_report(){
 copy_results() {
   local test_dir="$1"
   local all_result_dir="$2"
+  local test_script="${3:-test.sh}"
 
   local result_dir="${test_dir}/result"
-  local test_dir_name=$(basename ${test_dir})
-  if [[ -n "$(find "${result_dir}" -name "*.xml")" ]]; then
-    rebot --nostatusrc -N "${test_dir_name}" -l NONE -r NONE -o "${all_result_dir}/${test_dir_name}.xml" "${result_dir}/*.xml"
+  local test_dir_name="$(basename ${test_dir})"
+  local test_name="${test_dir_name}"
+  local target_dir="${all_result_dir}"/"${test_dir_name}"
+
+  if [[ -n "${test_script}" ]] && [[ "${test_script}" != "test.sh" ]]; then
+    local test_script_name=${test_script}
+    test_script_name=${test_script_name#test-}
+    test_script_name=${test_script_name#test_}
+    test_script_name=${test_script_name%.sh}
+    test_name="${test_name}-${test_script_name}"
+    target_dir="${target_dir}/${test_script_name}"
   fi
 
-  cp "${result_dir}"/docker-*.log "${all_result_dir}"/
-  if [[ -n "$(find "${result_dir}" -name "*.out")" ]]; then
-    cp "${result_dir}"/*.out* "${all_result_dir}"/
+  if [[ -n "$(find "${result_dir}" -name "*.xml")" ]]; then
+    rebot --nostatusrc -N "${test_name}" -l NONE -r NONE -o "${all_result_dir}/${test_name}.xml" "${result_dir}"/*.xml \
+      && rm -fv "${result_dir}"/*.xml "${result_dir}"/log.html "${result_dir}"/report.html
   fi
+
+  mkdir -p "${target_dir}"
+  mv -v "${result_dir}"/* "${target_dir}"/
 }
 
 run_test_script() {
   local d="$1"
-  local test_script="$2"
+  local test_script="${2:-test.sh}"
 
-  if [[ -z "$test_script" ]]; then
-    test_script=./test.sh
-  fi
-
-  echo "Executing test in ${d}"
+  echo "Executing test ${d}/${test_script}"
 
   #required to read the .env file from the right location
   cd "${d}" || return
 
   local ret=0
-  if ! "$test_script"; then
+  if ! ./"$test_script"; then
     ret=1
-    echo "ERROR: Test execution of ${d} is FAILED!!!!"
+    echo "ERROR: Test execution of ${d}/${test_script} is FAILED!!!!"
   fi
 
   cd - > /dev/null
@@ -355,15 +436,17 @@ run_test_script() {
 
 run_test_scripts() {
   local ret=0
+  local d f t
 
   for t in "$@"; do
     d="$(dirname "${t}")"
+    f="$(basename "${t}")"
 
-    if ! run_test_script "${d}"; then
+    if ! run_test_script "${d}" "${f}"; then
       ret=1
     fi
 
-    copy_results "${d}" "${ALL_RESULT_DIR}"
+    copy_results "${d}" "${ALL_RESULT_DIR}" "${f}"
 
     if [[ "${ret}" == "1" ]] && [[ "${FAIL_FAST:-}" == "true" ]]; then
       break
@@ -399,11 +482,95 @@ prepare_for_binary_image() {
 ##   (no binaries included)
 ## @param `ozone-runner` image version (optional)
 prepare_for_runner_image() {
-  local default_version=20210329-1 # set at build-time from Maven property
+  local default_version=20230615-1 # set at build-time from Maven property
   local runner_version=${OZONE_RUNNER_VERSION:-${default_version}} # may be specified by user running the test
+  local runner_image=${OZONE_RUNNER_IMAGE:-apache/ozone-runner} # may be specified by user running the test
   local v=${1:-${runner_version}} # prefer explicit argument
 
   export OZONE_DIR=/opt/hadoop
-  export OZONE_IMAGE="apache/ozone-runner:${v}"
+  export OZONE_IMAGE="${runner_image}:${v}"
 }
 
+## @description Executing the Ozone Debug CLI related robot tests
+execute_debug_tests() {
+  local prefix=${RANDOM}
+
+  local volume="cli-debug-volume${prefix}"
+  local bucket="cli-debug-bucket"
+  local key="testfile"
+
+  execute_robot_test ${SCM} -v "PREFIX:${prefix}" debug/ozone-debug-tests.robot
+
+  # get block locations for key
+  local chunkinfo="${key}-blocks-${prefix}"
+  docker-compose exec -T ${SCM} bash -c "ozone debug chunkinfo ${volume}/${bucket}/${key}" > "$chunkinfo"
+  local host="$(jq -r '.KeyLocations[0][0]["Datanode-HostName"]' ${chunkinfo})"
+  local container="${host%%.*}"
+
+  # corrupt the first block of key on one of the datanodes
+  local datafile="$(jq -r '.KeyLocations[0][0].Locations.files[0]' ${chunkinfo})"
+  docker exec "${container}" sed -i -e '1s/^/a/' "${datafile}"
+
+  execute_robot_test ${SCM} -v "PREFIX:${prefix}" -v "CORRUPT_DATANODE:${host}" debug/ozone-debug-corrupt-block.robot
+
+  docker stop "${container}"
+
+  wait_for_datanode "${container}" STALE 60
+  execute_robot_test ${SCM} -v "PREFIX:${prefix}" -v "STALE_DATANODE:${host}" debug/ozone-debug-stale-datanode.robot
+
+  wait_for_datanode "${container}" DEAD 60
+  execute_robot_test ${SCM} -v "PREFIX:${prefix}" debug/ozone-debug-dead-datanode.robot
+
+  docker start "${container}"
+
+  wait_for_datanode "${container}" HEALTHY 60
+}
+
+## @description  Wait for datanode state
+## @param        Datanode name, eg datanode_1 datanode_2
+## @param        State to check for
+## @param        The maximum time to wait in seconds
+wait_for_datanode() {
+  local datanode=$1
+  local state=$2
+  local timeout=$3
+
+  SECONDS=0
+  while [[ $SECONDS -lt $timeout ]]; do
+    local command="ozone admin datanode list"
+    docker-compose exec -T ${SCM} bash -c "$command" | grep -A2 "$datanode" > /tmp/dn_check
+    local health=$(grep -c "State: $state" /tmp/dn_check)
+
+    if [[ "$health" -eq 1 ]]; then
+      echo "$datanode is $state"
+      return
+    else
+      echo "Waiting for $datanode to be $state"
+    fi
+    echo "SECONDS: $SECONDS"
+  done
+  echo "WARNING: $datanode is still not $state"
+}
+
+
+## @description wait for n root certificates
+wait_for_root_certificate(){
+  local container=$1
+  local timeout=$2
+  local count=$3
+  local command="ozone admin cert list --role=scm -c 100 | grep -v "scm-sub" | grep "scm" | wc -l"
+
+  #Reset the timer
+  SECONDS=0
+  while [[ $SECONDS -lt $timeout ]]; do
+    cert_number=`docker-compose exec -T $container /bin/bash -c "$command"`
+    if [[ $cert_number -eq $count ]]; then
+      echo "$count root certificates are found"
+      return
+    fi
+      echo "$count root certificates are not found yet"
+      sleep 1
+  done
+  echo "Timed out waiting on $count root certificates. Current timestamp " $(date +"%T")
+  return 1
+}
